import textwrap


__all__ = ["format_h1", "format_h2", "fmt_error", "print_error", "menu", "format_progress", "markdown_table",
           "print_skipped", "format_exe_info", "format_box"]


NIND = 2  # Number of spaces per indentation level

# #################################################################################################
# # Text interface routines - routines that are useful for building a text interface


def format_underline(s, char="=", indents=0):
    """
    Traces a dashed line below string

    Args:
        s: string
        intents: number of leading intenting spaces
        format: string starting with "text" or "markdown"

    Returns: list

    >>> print("\\n".join(format_underline("Life of João da Silva", "^", 2)))
      Life of João da Silva
      ^^^^^^^^^^^^^^^^^^^^^
    """

    n = len(s)
    ind = " " * indents
    return ["{}{}".format(ind, s), "{}{}".format(ind, char*n)]


def format_h1(s, format="text", indents=0):
    """
    Encloses string in format text

    Args:
        s: string
        intents: number of leading intenting spaces
        format: string starting with "text" or "markdown"

    Returns: list

    >>> print("\\n".join(format_h2("Header 1", indents=10)))
              Header 1
              --------

    >>> print("\\n".join(format_h2("Header 1", "markdown", 0)))
    ## Header 1
    """

    if format.startswith("text"):
        return format_underline(s, "=", indents)
    elif format.startswith("markdown"):
        return ["# {}".format(s)]


def format_h2(s, format="text", indents=0):
    """
    Encloses string in format text

    Args:
        s: string
        format: string starting with "text" or "markdown"
        intents: number of leading intenting spaces

    Returns: list

    >>> print("\\n".join(format_h2("Header 2", indents=2)))
      Header 2
      --------

    >>> print("\\n".join(format_h2("Header 2", "markdown", 2)))
    ## Header 2
    """

    if format.startswith("text"):
        return format_underline(s, "-", indents)
    elif format.startswith("markdown"):
        return ["## {}".format(s)]



def fmt_error(s):
    """Standardized embellishment. Adds formatting to error message."""
    return "!! {0!s} !!".format(s)


def print_error(s):
    """Prints string as error message."""
    print((fmt_error(s)))


def menu(title, options, cancel_label="Cancel", flag_allow_empty=False, flag_cancel=True, ch='.'):
  """Text menu.

  Arguments:
    title -- menu title, to appear at the top
    options -- sequence of strings
    cancel_label='Cancel' -- label to show at last "zero" option
    flag_allow_empty=0 -- Whether to allow empty option
    flag_cancel=True -- whether there is a "0 - Cancel" option
    ch="." -- character to use to draw frame around title

  Returns:
    option -- an integer: None; 0-Back/Cancel/etc; 1, 2, ...

  Adapted from irootlab menu.m"""

  num_options, flag_ok = len(options), 0
  option = None  # result
  min_allowed = 0 if flag_cancel else 1  # minimum option value allowed (if option not empty)

  while True:
    print("")
    for line in format_box(title, ch):
        print("  "+line)
    for i, s in enumerate(options):
      print(("  {0:d} - {1!s}".format(i+1, s)))
    if flag_cancel: print(("  0 - << (*{0!s}*)".format(cancel_label)))
    try:
        s_option = input('? ')
    except KeyboardInterrupt:
        raise
    except:
        print("")

    n_try = 0
    while True:
      if n_try >= 10:
        print('You are messing up!')
        break

      if len(s_option) == 0 and flag_allow_empty:
        flag_ok = True
        break

      try:
        option = int(s_option)
        if min_allowed <= option <= num_options:
          flag_ok = True
          break
      except ValueError:
        print("Invalid integer value!")

      print(("Invalid option, range is [{0:d}, {1:d}]!".format(0 if flag_cancel else 1, num_options)))

      n_try += 1
      s_option = input("? ")

    if flag_ok:
      break
  return option


def format_box(title, ch="*"):
    """
    Encloses title in a box. Result is a list

    >>> for line in format_box("Today's TODO list"):
    ...     print(line)
    *************************
    *** Today's TODO list ***
    *************************
    """
    lt = len(title)
    return [(ch * (lt + 8)),
            (ch * 3 + " " + title + " " + ch * 3),
            (ch * (lt + 8))
           ]


def format_progress(i, n):
    """Returns string containing a progress bar, a percentage, etc."""
    if n == 0:
        fraction = 0
    else:
        fraction = float(i)/n
    LEN_BAR = 25
    num_plus = int(round(fraction*LEN_BAR))
    s_plus = '+'*num_plus
    s_point = '.'*(LEN_BAR-num_plus)
    return '[{0!s}{1!s}] {2:d}/{3:d} - {4:.1f}%'.format(s_plus, s_point, i, n, fraction*100)



def markdown_table(headers, data):
    """
    Creates MarkDown table. Returns list of strings

    Arguments:
      headers -- sequence of strings: (header0, header1, ...)
      data -- [(cell00, cell01, ...), (cell10, cell11, ...), ...]
    """

    maxx = [max([len(x) for x in column]) for column in zip(*data)]
    maxx = [max(ll) for ll in zip(maxx, [len(x) for x in headers])]
    mask = " | ".join(["%-{0:d}s".format(n) for n in maxx])

    ret = [mask % headers]



    ret.append(" | ".join(["-"*n for n in maxx]))
    for line in data:
        ret.append(mask % line)
    return ret


def print_skipped(reason):
    """Standardized printing for when a file was skipped."""
    print(("   ... SKIPPED ({0!s}).".format(reason)))



# #################################################################################################
# # Formatting for listing the programs available


def _format_exe_info(py_len, exeinfo, format, indlevel):
    """Renders ExeInfo object in specified format"""
    ret = []
    ind = " " * indlevel * NIND if format.startswith("text") else ""
    if format == "markdown-list":
        for si in exeinfo:
            ret.append("  - `{0!s}` -- {1!s}".format(si.filename, si.description))
    elif format == "markdown-table":
        mask = "%-{0:d}s | %s".format(py_len+2 )
        ret.append(mask % ("Script name", "Purpose"))
        ret.append("-" * (py_len + 3) + "|" + "-" * 10)
        for si in exeinfo:
            ret.append(mask % ("`{0!s}`".format(si.filename), si.description))
    elif format == "text":
        sbc = 1  # spaces between columns
        for si in exeinfo:
            ss = textwrap.wrap(si.description, 79 - py_len - sbc - indlevel*NIND)
            for i, s in enumerate(ss):
                if i == 0:
                    filecolumn = si.filename + " " + ("." * (py_len - len(si.filename)))
                else:
                    filecolumn = " " * (py_len + 1)

                ret.append("{}{}{}{}".format(ind, filecolumn, " "*sbc, s))
    ret.append("")
    return ret


def format_exe_info(exeinfo, format="text", indlevel=0):
    """
    Generates listing of all Python scripts available as command-line programs.

    Args:
      exeinfo -- list of ExeInfo objects

      format -- One of the options below:
        "text" -- generates plain text for printing at the console
        "markdown-list" -- generates MarkDown as a list
        "markdown-table" -- generates MarkDown as a table

      indents -- indentation level ("text" format only)

    Returns: (list of strings, maximum filename size)
      list of strings -- can be joined with a "\n"
      maximum filename size
    """

    py_len = max([len(si.filename) for si in exeinfo])

    sisi_gra = [si for si in exeinfo if si.flag_gui == True]
    sisi_cmd = [si for si in exeinfo if si.flag_gui == False]
    sisi_none = [si for si in exeinfo if si.flag_gui is None]

    def get_title(x):
        return format_h2(x, format, indlevel*NIND) + [""]

    ret = []
    if len(sisi_gra) > 0:
        ret.extend(get_title("Graphical applications"))
        ret.extend(_format_exe_info(py_len, sisi_gra, format, indlevel + 1))
    if len(sisi_cmd) > 0:
        ret.extend(get_title("Command-line tools", ))
        ret.extend(_format_exe_info(py_len, sisi_cmd, format, indlevel + 1))
    if len(sisi_none) > 0:
        ret.extend(_format_exe_info(py_len, sisi_none, format, indlevel + 1))

    return ret, py_len





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '.\XParametersEditor.ui'
#
# Created: Fri Jul 18 20:44:20 2014
#    by: PyQt5 UI code generator 4.10.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui

try:
  _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
  def _fromUtf8(s):
    return s

try:
  _encoding = QtGui.QApplication.UnicodeUTF8
  def _translate(context, text, disambig):
    return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
  def _translate(context, text, disambig):
    return QtGui.QApplication.translate(context, text, disambig)

class Ui_XParametersEditor(object):
  def setupUi(self, XParametersEditor):
    XParametersEditor.setObjectName(_fromUtf8("XParametersEditor"))
    XParametersEditor.setWindowModality(QtCore.Qt.WindowModal)
    XParametersEditor.resize(446, 62)
    self.horizontalLayout_2 = QtGui.QHBoxLayout(XParametersEditor)
    self.horizontalLayout_2.setSpacing(2)
    self.a99.set_margin(horizontalLayout_2, 2)
    self.horizontalLayout_2.setObjectName(_fromUtf8("horizontalLayout_2"))
    self.horizontalLayout = QtGui.QHBoxLayout()
    self.horizontalLayout.setSpacing(2)
    self.horizontalLayout.setObjectName(_fromUtf8("horizontalLayout"))
    self.frame = QtGui.QFrame(XParametersEditor)
    self.frame.setFrameShape(QtGui.QFrame.StyledPanel)
    self.frame.setFrameShadow(QtGui.QFrame.Raised)
    self.frame.setObjectName(_fromUtf8("frame"))
    self.horizontalLayout.addWidget(self.frame)
    self.buttonBox = QtGui.QDialogButtonBox(XParametersEditor)
    self.buttonBox.setOrientation(QtCore.Qt.Vertical)
    self.buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Cancel|QtGui.QDialogButtonBox.Ok)
    self.buttonBox.setObjectName(_fromUtf8("buttonBox"))
    self.horizontalLayout.addWidget(self.buttonBox)
    self.horizontalLayout_2.addLayout(self.horizontalLayout)

    self.retranslateUi(XParametersEditor)
    QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("rejected()")), XParametersEditor.reject)
    QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("accepted()")), XParametersEditor.accept)
    QtCore.QMetaObject.connectSlotsByName(XParametersEditor)

  def retranslateUi(self, XParametersEditor):
    XParametersEditor.setWindowTitle(_translate("XParametersEditor", "Parameters Editor", None))




"""Widget to choose a spectrum"""

# TODO get rid of this, maybe


__all__ = ["WChooseSpectrum"]

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99


class WChooseSpectrum(QWidget):
    """
    Editor to choose a spectrum: edit field with "..." button to find in directory

    Arguments:
      parent=None
    """

    # Emitted whenever any value changes
    edited = pyqtSignal()

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        # Whether all the values in the fields are valid or not
        self.flag_valid = False
        # Internal flag to prevent taking action when some field is updated programatically
        self.flag_process_changes = False
        self.sp = None # Spectrum object

        # # Central layout
        la = self.centralLayout = QHBoxLayout()
        a99.set_margin(la, 0)
        self.setLayout(la)

        y = self.lineEdit_sp = QLineEdit()
        la.addWidget(y)
        y.setReadOnly(True)
        y.installEventFilter(self)
        y.textEdited.connect(self.on_edited)

        b = self.b23088 = QPushButton("...")
        la.addWidget(b)
        b.clicked.connect(self.b_clicked)
        b.setFixedWidth(30)


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Interface

    def load(self, x):
        assert isinstance(x, a99.FileSpectrum)
        self.sp = x
        self.__update_from_sp()


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Qt override

    def setFocus(self, reason=None):
        pass

    def eventFilter(self, obj_focused, event):
        return False

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Slots

    def on_edited(self):
        pass
        # if self.flag_process_changes:
            # self.__update_f()
            # self.edited.emit()

    def b_clicked(self):
        # index = self.__get_index()
        # editor, text, cls, label, wild = self.editors[index], \
        #                                  self.open_texts[index], self.clss[index], self.labels_fn[index], \
        #                                  self.wilds[index]
        try:
            d = self.sp.filename if self.sp and self.sp.filename is not None else a99.FileSpectrumPfant.default_filename
            new_filename = QFileDialog.getOpenFileName(self, "Cube", d, "*.fits")[0]
            if new_filename:
                sp = a99.load_spectrum(new_filename)
                if not sp:
                    raise RuntimeError("Failed to load '{0!s}'".format(new_filename))
                self.sp = sp
                self.__update_from_sp()
        except Exception as e:
            a99.show_error(str(e))
            raise


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Internal gear

    def __update_from_sp(self):
        self.flag_process_changes = False
        try:
            self.lineEdit_sp.setText(self.sp.filename)
        finally:
            self.flag_process_changes = True



import datetime
import os
import traceback as tb

from PyQt5.QtCore import Qt
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from matplotlib import pyplot as plt
import a99
from .a_XLogMainWindow import *


class XFileMainWindow(XLogMainWindow):
    """Application template with file operations in two tabs: (application area) and (log)"""

    def __init__(self, parent=None):
        def keep_ref(obj):
            self._refs.append(obj)
            return obj

        XLogMainWindow.__init__(self, parent)
        self._refs = []

        # This may become useful in the future, but for now...
        # # XRunnableManager instance
        # self._manager_form = None
        # # RunnableManager instance
        # self._rm = None

        self.save_dir = None
        self.load_dir = None

        # # Synchronized sequences
        # Replace first element of each attribute at child class
        self.tab_texts =  ["------ (Alt+&1)", "Log (Alt+&2)"]
        self.flags_changed = [False, False]
        self.save_as_texts = [None, None]
        self.open_texts = [None, None]
        self.clss = [None, None]  # save class
        self.clsss = [None, None]  # accepted load classes
        self.wilds = [None, None]  # e.g. '*.fits'
        self.editors = [NullEditor(), NullEditor()]  # editor widgets, must comply ...

        # # Menu bar
        b = self.menuBar()
        m = self.menu_file = b.addMenu("&File")

        ac = m.addAction(a99.get_icon("document-open"), "&Open...")
        ac.setShortcut("Ctrl+O")
        ac.triggered.connect(self.on_open)

        m.addSeparator()

        ac = m.addAction(a99.get_icon("document-save"), "&Save")
        ac.setShortcut("Ctrl+S")
        ac.triggered.connect(self.on_save)

        ac = m.addAction(a99.get_icon("document-save-as"), "Save &as...")
        ac.setShortcut("Ctrl+Shift+S")
        ac.triggered.connect(self.on_save_as)

        ac = m.addAction("Save a&ll...")
        ac.setShortcut("Alt+Shift+S")
        ac.triggered.connect(self.on_save_all)

        m.addSeparator()

        ac = m.addAction("Load &default")
        ac.setStatusTip("Loads default")
        ac.setShortcut("Ctrl+D")
        ac.triggered.connect(self.on_reset)

        m.addSeparator()

        ac = m.addAction(a99.get_icon("system-shutdown"), "&Quit")
        ac.setShortcut("Ctrl+Q")
        ac.triggered.connect(self.close)


        # # Central layout
        ##################
        cw = self.centralWidget = QWidget()
        self.setCentralWidget(cw)
        lantanide = self.centralLayout = QVBoxLayout(cw)
        a99.set_margin(lantanide, 1)

        # ## Tabs
        tt = self.tabWidget = QTabWidget(self)
        lantanide.addWidget(tt)
        tt.setFont(a99.MONO_FONT)
        # ### Custom Area tab will have a widget **without** a layout, named "gotting"
        w0 = self.gotting = QWidget()
        tt.addTab(w0, self.tab_texts[0])

        # ### Log tab
        te = self.textEdit_log = keep_ref(QTextEdit())
        te.setReadOnly(True)
        tt.addTab(te, self.tab_texts[1])

        # ### Final tabs setup
        tt.setCurrentIndex(0)

        # ## "Last Log" label
        x = self.label_last_log = QLabel(self)
        lantanide.addWidget(x)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Interface

    def load(self, f):
        index = self._get_tab_index()
        if not isinstance(f, self.clsss[index]):
            raise RuntimeError('Object to load must be in {0!s} (not a {1!s})'.format([x.__name__ for x in self.clsss[index]], f.__class__.__name__))
        f = self._filter_on_load(f)

        assert f.filename is not None, "_filter_on_load() probably forgot to assign file name"

        editor = self.editors[index]
        editor.load(f)
        self._update_tab_texts()


    def keep_ref(self, obj):
        """Adds obj to internal list to keep a reference to it.

        WHen using PyQt, it happens that the Python object gets garbage-collected even
        when a C++ Qt object still exists, causing a mess
        """
        self._refs.append(obj)
        return obj

    def add_log_error(self, x, flag_also_show=False):
        """Sets text of labelError."""
        if len(x) == 0:
            x = "(empty error)"
            tb.print_stack()
        x_ = x
        x = '<span style="color: {0!s}">{1!s}</span>'.format(a99.COLOR_ERROR, x)
        if hasattr(self, "label_last_log"):
            self.label_last_log.setText(x)
        self.add_log(x, False)
        if flag_also_show:
            a99.show_error(x_)


    def add_log(self, x, flag_also_show=False):
        """Sets text of labelDescr."""
        if hasattr(self, "textEdit_log"):
            te = self.textEdit_log
            te.append("{0!s} -- {1!s}".format(datetime.datetime.strftime(datetime.datetime.now(), '%Y-%m-%d %H:%M:%S'), x))
        if hasattr(self, "label_last_log"):
            self.label_last_log.setText(x)

        a99.get_python_logger().info(x)
        if flag_also_show:
            a99.show_error(x)


    # def set_manager_form(self, x):
    #     assert isinstance(x, XRunnableManager)
    #     self._manager_form = x
    #     self._rm = x.rm


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Qt override

    def closeEvent(self, event):
        flag_exit, ff = True, []
        for ed, flag_changed in zip(self.editors, self.flags_changed):
            if ed and ed.f and flag_changed:
                ff.append(ed.f.description)

        if len(ff) > 0:
            s = "Unsaved changes\n  -" + ("\n  -".join(ff)) + "\n\nAre you sure you want to exit?"
            flag_exit = a99.are_you_sure(True, event, self, "Unsaved changes", s)
        if flag_exit:
            plt.close("all")


    def keyPressEvent(self, evt):
        incr = 0
        if evt.modifiers() == Qt.ControlModifier:
            n = self.tabWidget.count()
            if evt.key() in [Qt.Key_PageUp, Qt.Key_Backtab]:
                incr = -1
            elif evt.key() in [Qt.Key_PageDown, Qt.Key_Tab]:
                incr = 1
            if incr != 0:
                new_index = self._get_tab_index() + incr
                if new_index < 0:
                    new_index = n - 1
                elif new_index >= n:
                    new_index = 0
                self.tabWidget.setCurrentIndex(new_index)


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Slots for Qt library signals

    # All file operations

    def on_open(self):
        if self._tab_has_file_operations():
            self.__generic_open()

    def on_save(self):
        if self._tab_has_file_operations():
            self.__generic_save()

    def on_save_as(self):
        if self._tab_has_file_operations():
            self.__generic_save_as()

    def on_save_all(self):
        index_save = self._get_tab_index()
        try:
            for index in range(self.tabWidget.count()):
                self.tabWidget.setCurrentIndex(index)
                if not self.__generic_save():
                    # breaks if user cancels a "save as" operation
                    break
        finally:
            self.tabWidget.setCurrentIndex(index_save)

    def on_reset(self):
        if self._tab_has_file_operations():
            idx = self._get_tab_index()
            editor = self.editors[idx]
            flag_ok = True
            if editor.f:
                descr = self._get_tab_description()
                r = QMessageBox.question(self, "Load default", "Current setup "
                                                               "for %s will be overwritten with a 'default' setup.\n\n"
                                                               "Confirm?" % descr, QMessageBox.Yes | QMessageBox.No,
                                         QMessageBox.Yes)
                flag_ok = r == QMessageBox.Yes
            if flag_ok:
                self._generic_reset()


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Protected methods to be overriden or used by descendant classes

    def _on_edited(self):
        index = self._get_tab_index()
        self.flags_changed[index] = True
        self._update_tab_texts()



    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Protected methods to override as needed

    def _filter_on_load(self, f):
        """
        Inherit this function to convert file type

        The idea is to handle multiple file types on load. If "f" class is not supported,
        this method must raise an Exception
        """
        return f

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Gear

    def _get_tab_index(self):
        """Returns index of current selected tab."""
        return self.tabWidget.currentIndex()

    def _get_tab_description(self):
        """Returns "description" of current tab (tab text without shortcut info)."""
        idx = self._get_tab_index()
        text = self.tab_texts[idx]
        if "(" in text:
            text = text[:text.index("(") - 1]
        text = text[0].lower() + text[1:]
        return text

    def _tab_has_file_operations(self):
        return self.flags_changed[self._get_tab_index()] is not None

    def _update_tab_texts(self):
        for index, (text, flag_changed) in enumerate(zip(self.tab_texts, self.flags_changed)):
            self.tabWidget.setTabText(index,
                                      text + (" (changed)" if flag_changed else ""))

    def _generic_reset(self):
        index = self._get_tab_index()
        editor, text, cls = self.editors[index], self.open_texts[index], \
                            self.clss[index]
        f = cls()
        f.init_default()
        editor.load(f)
        # perhaps it is save to assume that the status is not 'changed' if there is a filename
        # assigned after init_default, but TODO I should implement a class (instead of horrible synchronized sequences) and set flag_changeable to False... someday
        self.flags_changed[index] = f.filename is None
        self._update_tab_texts()

    def __generic_save(self):
        """Returns False if user has cancelled a "save as" operation, otherwise True."""
        index = self._get_tab_index()
        editor = self.editors[index]
        f = editor.f
        if not f:
            return True
        if not editor.flag_valid:
            a99.show_error("Cannot save, {0!s} has error(s)!".format(f.description))
            return True
        if f.filename:
            try:
                f.save_as()
                self.flags_changed[index] = False
                self._update_tab_texts()
                editor.update_gui_label_fn()  # duck typing
                return True
            except Exception as e:
                a99.show_error(str(e))
                raise
        else:
            return self.__generic_save_as()

    def __generic_save_as(self):
        """Returns False if user has cancelled operation, otherwise True."""
        index = self._get_tab_index()
        editor, text, wild = self.editors[index], self.save_as_texts[index], \
                             self.wilds[index]
        if not editor.f:
            return True
        if editor.f.filename:
            d = editor.f.filename
        else:
            d = os.path.join(self.save_dir if self.save_dir is not None \
                                 else self.load_dir if self.load_dir is not None \
                else ".", editor.f.default_filename)
        new_filename = QFileDialog.getSaveFileName(self, text, d, wild)[0]
        if new_filename:
            self.save_dir, _ = os.path.split(str(new_filename))
            try:
                editor.f.save_as(str(new_filename))
                self.flags_changed[index] = False
                self._update_tab_texts()
                editor.update_gui_label_fn()
                return True
            except Exception as e:
                a99.show_error(str(e))
                raise
        return False

    def __generic_open(self):
        index = self._get_tab_index()
        editor, text, cls, wild = self.editors[index], \
                                  self.open_texts[index], self.clss[index], self.wilds[index]
        try:
            d = self.load_dir if self.load_dir is not None \
                else self.save_dir if self.save_dir is not None \
                else "."
            new_filename = QFileDialog.getOpenFileName(self, text, d, wild+";;*.*")[0]
            if new_filename:
                self.load_filename(new_filename)
        except Exception as e:
            a99.show_error(str(e))
            raise

    def load_filename(self, filename, index=None):
        """
        Loads file given filename

        Arguments:
            filename --
            index -- tab index to load file into. If not passed, loads into current tab
        """
        filename = str(filename)  # QString protection
        if index is None:
            index = self._get_tab_index()
        editor, text, cls, wild, possible_classes = self.editors[index], \
                                                    self.open_texts[index], self.clss[index], self.wilds[index], \
                                                    self.clsss[index]
        self.load_dir, _ = os.path.split(filename)
        f = a99.load_with_classes(filename, possible_classes)
        if f is None:
            raise RuntimeError("Could not load '{0!s}'".format(filename))

        f = self._filter_on_load(f)

        self.load(f)


class NullEditor(object):
    """Class to fulfill requirement in XFileMainWindow, easiest way out, no bother"""
    f = None